//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* file_in.cpp                                                            *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://www.messen-und-deuten.de/frida                                  *//
//**************************************************************************//

#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <map>

#include <ask_simple_value.h>
#include <yaml-cpp.h>

#include "mystd.h"
#include "olf.h"
#include "mem.h"
#include "var.h"
#include "expr.h"
#include "file_in.h"

using namespace std;

namespace NFileIn {
    void Load_08( ifstream& F_in, string flong );
    void Load_01( FILE *F_in, string flong );
    void LoadSpec_01( FILE *F_in, PSpec& sout, int nz );
    void Load_96( FILE *F_in, string flong );
}


//! Load file in current or legacy format.

void NFileIn::Load(void)
{
    string fnames = sask( "Load file(s)" );
    vector<string> vflong; // unix names
    mystd::glob_file_list( fnames, "y08", &vflong );
    FILE *F_in;
    // cout << "DEBUG: size " << vflong.size() << "\n";
    for( uint i=0; i<vflong.size(); ++i ) {
        if( !(F_in = fopen(vflong[i].c_str(), "r")) )
            throw string( "cannot open file " ) + vflong[i];
        try{ 
            cout << ".. loading file " << vflong[i] << "\n";
            string fdir, fshort, fext;
            mystd::fname_divide( vflong[i], &fdir, &fshort, &fext);
            
            if        ( fext=="y08" ) {
                ifstream FS (vflong[i].c_str(), ios_base::in);
                if( FS.fail() )
                    throw "cannot open file " + vflong[i];
                Load_08( FS, vflong[i] );
            	FS.close();
            } else if ( fext=="a01" ) {
                Load_01( F_in, vflong[i] );
            } else if ( fext=="i96" ) {
                Load_96( F_in, vflong[i] );
            } else {
                throw "unknown file extension [" + fext + "]";
            }
            
            fclose( F_in );
        }
        catch(exception& e) {
            cout << "DEBUG: catched locally\n";
            cerr << e.what() << endl;
            if (F_in) fclose(F_in);
        }
        catch(string& s) {
            // for y08 only
            cerr << s << endl;
            if (F_in) fclose(F_in);
        }
    }
}


//! Load a YAML file in y08 format.

void NFileIn::Load_08( ifstream& FS, string flong )
{
    string lin, key, val, fdir, fshort, fext, res;
    double num, vx, vy, dy;
    uint iz;
    bool isdata;
    CCoord co;

    mystd::fname_divide( flong, &fdir, &fshort, &fext);

    YAML::Parser parser(FS);
    YAML::Node doc;
    parser.GetNextDocument(doc);

    // read meta
    if(!doc.FindValue("Meta"))
        throw string("no Meta" );
    try {
        YAML::CONTENT_TYPE Metatype = doc["Meta"].GetType();
        if ( Metatype != YAML::CT_MAP && doc["Meta"].size() > 0)
            throw string("Meta is not a MAP type" );
        if(!doc["Meta"].FindValue("format"))
            throw string("no format in Meta" );
        string str = doc["Meta"]["format"];
        if ( str != "frida/y08 for yaml1" )
            throw string("format is not frida/y08 for yaml1" );
    } catch(YAML::ParserException& e) {
    	std::cout << "no Meta" << e.what() << "\n";
    }
    if(!doc["Meta"].FindValue("type"))
        throw string("no type in Meta" );
    string type =  doc["Meta"]["type"];

    // create output file
    POlo fout;
    POld fd;
    POlc fc;
    if ( type == "generic tabular data" ){
        isdata = true;
        fd = POld( new COld );
        fout = fd;
    } else if ( type == "frida2 curve" ){
        isdata = false;
        fc = POlc( new COlc );
        fout = fc;
    } else
        throw "File "+flong+" has invalid type "+type;
    
    // read history
    if(!doc.FindValue("History"))
        throw string("no History" );
    if ( doc["History"].GetType() != YAML::CT_SEQUENCE  &&
         doc["History"].size() > 0 )
        throw string("History is not a SEQUENCE type" );
    for (unsigned int i = 0; i < doc["History"].size(); i++) {
        string str;
        doc["History"][i] >> str;
        fout->lDoc.push_back(str);
    }

    // read coord	
    if(!doc.FindValue("Coord"))
        throw string("no Coord" );
    if ( doc["Coord"].GetType() != YAML::CT_MAP && doc["Coord"].size() > 0 )
        throw string("Coord is not a MAP type" );
    if( const YAML::Node *pName =  doc["Coord"].FindValue("x") ){
        (*pName)["name"] >> fout->xco.name;
        (*pName)["unit"] >> fout->xco.unit;
    } else
        throw string( "no x coord" );
    if( const YAML::Node *pName =  doc["Coord"].FindValue("y") ){
        (*pName)["name"] >> fout->yco.name;
        (*pName)["unit"] >> fout->yco.unit;
    } else
        throw string( "no y coord" );
   
    iz=0;
    while(const YAML::Node *pName =  doc["Coord"].FindValue("z"+strg(iz))){
        (*pName)["name"] >> co.name;
        (*pName)["unit"] >> co.unit;
      	fout->ZCo.push_back( co );
    	++iz;
    }    	

    // read Param	
    if(!doc.FindValue("Param"))
        throw string("no Param" );
    if ( doc["Param"].GetType() != YAML::CT_SEQUENCE  &&
         doc["Param"].size() > 0)
        throw string("Param is not a SEQUENCE type" );
    for(unsigned int iParam = 0; iParam < (doc["Param"].size()); iParam++ ){
        if ( doc["Param"][iParam].GetType() != YAML::CT_MAP )
            throw "Param " + strg(iParam) + " is not a MAP type";
        doc["Param"][iParam]["name"] >> co.name;
        doc["Param"][iParam]["unit"] >> co.unit;
        doc["Param"][iParam]["value"] >> val;
        if( !mystd::any2dbl( val, &num ) )
            throw string( "param(" ) + co.name + "): invalid value " + val;
        fout->RPar.push_back( CParam( co, num ) );
    }
	
    // for curve, read formula
    if( !isdata ) {
        const YAML::Node *pName = doc.FindValue("Formula");
        if( !pName )
            throw string( "DEFICIENT FILE: no formula" );
        string expr;
        *pName >> expr;
        fc->parseFunction( expr );
        uint np = fc->T->npar();
        for( uint ip=0; ip<np; ip++ )
            fc->PCo.push_back( CCoord("p" + strg(ip), "") );
        fc->setAux();
    }

    //start to read Tables
    if(!doc.FindValue("Tables"))
        throw string("no Tables" );
    if ( doc["Tables"].GetType() != YAML::CT_SEQUENCE &&
         doc["Tables"].size() > 0 )
        throw string("Tables is not a SEQUENCE type" );
    for (unsigned int iTable = 0; iTable < (doc["Tables"].size()); iTable++) {	
        if ( doc["Tables"][iTable].GetType() != YAML::CT_MAP )
            throw string("Tables " + strg(iTable) + " is not a MAP type" );
        vector<double> z;
        for(iz=0; iz<fout->ZCo.size(); ++iz ){
            doc["Tables"][iTable]["z" + strg(iz)] >> val;
            if( !mystd::any2dbl( val, &num ) )
                throw "z" + strg(iz) + ": invalid value " + val;
            z.push_back( num );
        }

        if( isdata ) {
            PSpec sout( new CSpec );
            sout->z = z;
            YAML::Iterator itxy=doc["Tables"][iTable].begin();
            string xytag = itxy.first();
            if ( xytag == "xy" ){
                itxy.second() >> val;
                while( val != "" ) {
                    mystd::string_extract_line( val, &lin, &res );
                    if( sscanf( lin.c_str(), "%lg %lg\n", &vx, &vy )!=2 )
                       	throw "xy: bad data line: " + lin;
                    sout->push_xy( vx, vy );
                    val = res;
               	};
            }else if ( xytag == "xyd" ){
                itxy.second() >> val;
               	while( val != "" ) {
                    mystd::string_extract_line( val, &lin, &res );
                    if( sscanf( lin.c_str(),
                                "%lg %lg %lg\n",&vx, &vy, &dy )!=3 )
                       	throw "xyd: bad data line: " + lin;
                    sout->push_xyd( vx, vy, dy );
                    val = res;
               	};
            } else
                throw "invalid xytag " + xytag;
            fout->V.push_back( sout );
        } else {
            PCurve cout( new CCurve );
            cout->z = z;
            for( uint ip=0; ip<fc->nPar(); ++ip ){
                doc["Tables"][iTable]["p" + strg(ip)] >> val;
                if( !mystd::any2dbl( val, &num ) )
                    throw "p" + strg(ip) + ": invalid value " + val;
                cout->P.push_back( num );
                cout->fixed.push_back( false );
            }
            fout->V.push_back( cout );
        }
    }

    fout->name = fshort;
    fout->lDoc.push_back( "loaded as " + flong );
    fout->as_on_disk = true;
    NOlm::OloAdd( fout );
}


//! Load a .a01 formatted file.

void NFileIn::Load_01( FILE *F_in, string flong )
{
    string lin, key, val, fdir, fshort, fext;
    uint iz, nj=0, j;
    CCoord co;
    POld fout( new COld );

    mystd::fname_divide( flong, &fdir, &fshort, &fext);

    if     ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "file empty or starting with empty line" );
    if     ( lin!=string("ASCII-01 JWu") )
        throw ".a01 file starts with illegal header [" + lin + "]";

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<5 || 
         lin.substr(0,4)!=string("typ ") )
        throw string( "no valid typ line" );
    if        ( lin.substr(4)=="D" ) {
        // ok
    } else if ( lin.substr(4)=="C" ) {
        throw string( "reading curve files not implemented" );
    } else {
        throw string( "invalid type [" + lin.substr(4) + "]" );
    }

    while (1) { // loop over input lines

        if        ( mystd::freadln(F_in, &lin)<0 )
            throw string( "unexpected end-of-file" );

        // regular exit when "eof" is reached:
        if        (lin=="eof") {
            fout->as_on_disk = true;
            NOlm::OloAdd( fout );
            return;
        }

        if        ( lin.size() < 4 )
            throw string( "incomplete line" );

        key = lin.substr(0,4);
        val = lin.substr(4);

        if        (key=="fil ") {
            if (val!=fshort) {
                cout << "warning: file '" << fshort <<
                    "' renamed from '" << val << "'\n";
            }
            fout->name = fshort;
        } else if (key=="doc ") {
            fout->lDoc.push_back(val);
        } else if (key=="x   ") {
            fout->xco.load_a01(val);
        } else if (key=="y   ") {
            fout->yco.load_a01(val);
        } else if (lin[0]=='z') {
            if ( sscanf(lin.substr(1,3).c_str(), "%u", &iz)!=1 )
                throw string( "z coordinates must be numbered" );
            if ( iz!=fout->ZCo.size() )
                throw string( "z coordinates must be sorted" );
            co.load_a01(val);
            fout->ZCo.push_back(co);
        } else if (lin.substr(0,4)=="rpa ") {
            string in = lin.substr(4);
            int ji = in.find_first_not_of(" \t");
            double dval;
            if( ji>=20 || !mystd::any2dbl( in.substr(ji,20), &dval ) )
                throw string( "invalid rpa line\n" );
            fout->RPar.push_back( CParam( CCoord(), dval ) );
            throw string( "a01 file contains rpa section. Check the source code preceeding this error message. It is possibly broken. If not, remove this exception.\n" );
        } else if (key=="#j  ") {
            if ( sscanf(val.c_str(), "%u", &nj)!=1 )
               throw string( "after key #j no no. of spectra\n");
        } else if (key=="scn ") {
            if ( sscanf(val.c_str(), "%u", &j)!=1 ) {
                cout << "warning: number of spectra not given\n";
            } else if ( j!=fout->nJ() ) {
                cout << "warning: spec " << fout->nJ() <<
                    " has label " << j << "\n";
            }
            PSpec sout( new CSpec );
            LoadSpec_01( F_in, sout, fout->nZ() );
            fout->V.push_back( sout );
        } else {
            throw "Unexpected key [" + key + "]";
        }
    }
}


//! Load a spectrum from a .a01 formatted files.

void NFileIn::LoadSpec_01( FILE *F_in, PSpec& sout, int nz )
{
    int err;
    uint i, m, n;
    char lab[4];
    double v, w;

    // get z values:
    if ((err=fscanf(F_in, "%4c%u\n", lab, &m))!=2)
        throw "expecting Z head line, found " + strg(err) + " entries";
    if (strncmp(lab, "Z   ", 4))
        throw "found [" + string(lab) + "] in Z head line";
    if (m!=nz)
        throw "found " + strg(m) + " Z entries, expecting " + strg(nz);
    for ( i=0; i<nz; ++i ) {
        if ((err=fscanf(F_in, "%lg\n", &v))!=1)
            throw "expecting z value, found " + strg(err) + " entries";
        sout->z.push_back( v );
    }

    // get xy values:
    if ( (err=fscanf(F_in, "%4c%u\n", lab, &n))!=2 )
        throw "expecting XY head line, found " + strg(err) + " entries";
    if (strncmp(lab, "XY  ", 4) )
        throw "found [" + string(lab) + "] in XY head line";
    
    for (i=0; i<n; i++) {
        if ((err=fscanf(F_in, "%lg %lg\n", &v, &w))!=2) {
            throw string( "expecting xy pair, found strg(err) entries" );
        }
        sout->push_xy(v, w);
    }
}


//! Load a .i96 formatted file.

void NFileIn::Load_96( FILE *F_in, string flong )
{
    string lin, key, val, fdir, fshort, fext;
    double r0, r1, r2;
    uint n, nspec, ival;
    POld fout( new COld );
    CCoord co;
    PSpec sout;

    mystd::fname_divide( flong, &fdir, &fshort, &fext);
    fout->name = fshort;

    if     ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "file empty or starting with empty line" );
    if     ( lin!=string("ASCII-96") )
        throw ".i96 file starts with illegal header [" + lin + "]";

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<9 || 
         lin.substr(0,9)!=string("(a24,a56)") )
        throw string( "no valid intro to block 2" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<24 ||
         lin.substr(0,3)!=string("fil") )
        throw string( "tag 'fil' not found in block 2" );
    if ( mystd::strip(lin.substr(24))!=fshort ) {
                cout << "warning: i96 file '" << fshort <<
                    "' renamed from '" << lin.substr(24) << "'\n";
    }

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<24 ||
         lin.substr(0,3)!=string("tit") )
        throw string( "tag 'tit' not found in block 2" );
    fout->lDoc.push_back( mystd::strip(lin.substr(24)) );

    while( mystd::freadln(F_in, &lin)>0 && lin.length()>=24 &&
         lin.substr(0,3)==string("doc") )
        fout->lDoc.push_back( mystd::strip(lin.substr(24)) );

    if( lin.substr(0,3)!=string("dir") )
        throw string( "tag 'dir' not found in block 2" );

    while ( mystd::freadln(F_in, &lin)>0 &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 2") ) )
        ;
    
    if ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "missed &eob 3" );

    if( lin.length()<9 || lin.substr(0,9)!=string("(a24,i16)") )
        throw string( "no valid intro to block 3" );

    while ( mystd::freadln(F_in, &lin)>0 &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 3") ) )
        ;
    
    if ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "missed &eob 3" );

    if ( lin.length()<16 || lin.substr(0,16)!=string("(a24,a24,g20.10)") )
        throw string( "no valid intro to block 4" );

    while ( mystd::freadln(F_in, &lin)>0 && lin.length()>=48 &&
            lin.substr(0,4)!=string("&eob") ){
        if( sscanf( lin.substr(48).c_str(), "%lg", &r0 )!=1 )
            throw string( "no real value in rpar line" );
        fout->RPar.push_back(
            CParam( lin.substr(0,24), lin.substr(24,24), r0 ) );
    }

    if( lin.substr(0,6)!=string("&eob 4") )
        throw string( "no valid eob 4" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<9 || 
         lin.substr(0,12)!=string("(a4,a24,a24)") )
        throw string( "no valid intro to block 5" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<28 || 
         lin.substr(0,1)!=string("x") )
        throw string( "no x coordinate" );
    fout->xco = CCoord( mystd::strip(lin.substr(4,24)),
                        mystd::strip(lin.substr(28,24)) );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<28 || 
         lin.substr(0,1)!=string("y") )
        throw string( "no y coordinate" );
    fout->yco = CCoord( mystd::strip(lin.substr(4,24)),
                        mystd::strip(lin.substr(28,24)) );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<28 || 
         lin.substr(0,1)!=string("z") )
        throw string( "no z coordinate" );
    fout->ZCo.push_back( CCoord( mystd::strip(lin.substr(4,24)),
                                 mystd::strip(lin.substr(28,24)) ) );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<6 || 
         lin.substr(0,6)!=string("&eob 5") )
        throw string( "no valid eob 5" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<5 || 
         lin.substr(0,5)!=string("(a80)") )
        throw string( "no valid intro to block 6" );

    while ( mystd::freadln(F_in, &lin)>0 &&
            !( lin.length()>=6 &&
               lin.substr(0,6)==string("&eob 6") ) )
        fout->lDoc.push_back( lin );

    if ( mystd::freadln(F_in, &lin)<=0 )
        throw string( "missed &eob 6" );
    for( int i=1; i<3; ++i )
        if ( mystd::freadln(F_in, &lin)<=0 )
            throw string( "no valid intro to block 7" );

    if ( mystd::freadln(F_in, &lin)<=0 || 
         sscanf( lin.c_str(), "%i %i", &nspec, &ival )!=2 )
            throw string( "no nspec or 2nd arg in intro to block 7" );

    for( uint j=0; j<nspec; ++j ){

        sout = PSpec( new CSpec );

        if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<10 || 
             lin.substr(0,9)!=string("&spectrum") ||
             sscanf( lin.substr(10).c_str(), "%i", &ival )!=1 ||
             ival!=j+1 )
            throw string( "no valid intro to spectrum" );

        if ( mystd::freadln(F_in, &lin)<0 ||
             sscanf( lin.c_str(), "%i %lg", &n, &r0 )!=2 )
            throw string( "no valid header in spectrum" );

        sout->z.push_back( r0 );

        for( uint i=0; i<n; ++i ){
            if ( !F_in || mystd::freadln(F_in, &lin)<=0 )
                throw "i96: failed to read data line " +
                    strg(i) + " of " + strg(n);
            if ( sscanf( lin.c_str(), "%lg %lg %lg", &r0, &r1, &r2 )!=3 )
                throw "i96: bad data line: '" + lin + "'";
            sout->push_xy(r0, r1);
        }

        fout->V.push_back( sout );

    }

    fout->as_on_disk = true;
    NOlm::OloAdd( fout );

}

